@@ -1,4 +1,4 @@
-from typing import Optional, Union, Literal
+from typing import Optional, Union, Literal, Dict, Set
 from pathlib import Path
 from pydantic import Extra
 from pydantic_yaml import YamlModel
@@ -6,6 +6,7 @@ import logging
 import tempfile
 import click
 import re
+import json
 
 # unfortunately there is no type info ATM
 import graphviz  # type: ignore
@@ -24,7 +25,7 @@ class URL(YamlModel):
 class Assignment(YamlModel):
     description: str
     path: Path
-    attachments: Optional[list[Path]]
+    attachments: Optional[tuple[Path, ...]]
 
     class Config:
         extra = Extra.forbid
@@ -53,9 +54,12 @@ class Node(YamlModel, extra=Extra.forbid):
 
     id: str
     title: Optional[str]
-    assignments: Optional[list[Assignment]]
-    urls: Optional[list[URL]]
-    jobs: Optional[list[Job]]
+    assignments: Optional[tuple[Assignment, ...]]
+    urls: Optional[tuple[URL, ...]]
+    jobs: Optional[tuple[Job, ...]]
+
+    def __hash__(self):
+        return hash(self.id)
 
 
 class NodePlaceholder(YamlModel, extra=Extra.forbid):
@@ -104,32 +108,16 @@ def visualize_module(paths) -> None:
     module = Module(clusters=[])
     for path in paths:
         module.clusters.append(Cluster.parse_file(path))
+    nxgraph = _module_dependency_graph(module)
     dot = graphviz.Digraph()
-    for cluster in module.clusters:
-        nodes: list[Node] = cluster.nodes
-        for node in nodes:
-            dot.node(
-                f"{cluster.namespace_prefix}__{node.id}",
-                f"({cluster.namespace_prefix}) {node.title or node.id}",
-            )
-        edges: list[Edge] = cluster.edges
-        for edge in edges:
-            start_id = edge.start_id
-            end_id = edge.end_id
-            if "__" not in start_id:
-                start_id = f"{cluster.namespace_prefix}__{start_id}"
-            if "__" not in end_id:
-                end_id = f"{cluster.namespace_prefix}__{end_id}"
-            dot.edge(start_id, end_id)
-        if cluster.motivations:
-            for motivation in cluster.motivations:
-                start_id = motivation.start_id
-                end_id = motivation.end_id
-                if "__" not in start_id:
-                    start_id = f"{cluster.namespace_prefix}__{start_id}"
-                if "__" not in end_id:
-                    end_id = f"{cluster.namespace_prefix}__{end_id}"
-                dot.edge(start_id, end_id, style="dotted")
+    for node in nxgraph.nodes:
+        dot.node(node.id, node.title or node.id)
+    for (u, v) in nxgraph.edges:
+        kind = nxgraph.get_edge_data(u, v)["kind"]
+        if kind == "dependency":
+            dot.edge(u.id, v.id)
+        elif kind == "motivation":
+            dot.edge(u.id, v.id, style="dotted")
     dot.render("module.gv", directory=tempfile.gettempdir(), cleanup=True, view=True)
 
 
@@ -140,6 +128,7 @@ cli.add_command(visualize_module)
 @click.argument("path", type=click.Path(exists=True))
 def visualize_cluster(path) -> None:
     """Show a visual representation of a cluster, with minimal namespacing."""
+    # does not seem worthwhile rewriting with networkx
     cluster: Cluster = Cluster.parse_file(path)
     dot = graphviz.Digraph()
     nodes: list[Node] = cluster.nodes
@@ -174,10 +163,6 @@ def writeschema() -> None:
         fh.write(Cluster.schema_json(indent=2))
 
 
-def check_redundancy(module: Module) -> Optional[str]:
-    raise NotImplementedError("Cannot check for redundancy yet.")
-
-
 @click.command()
 @click.argument("paths", type=click.Path(exists=True), nargs=-1)
 def validate_module(paths) -> None:
@@ -260,7 +245,7 @@ def validate_module(paths) -> None:
     for message in messages:
         print(message)
     messages = []
-    module_dag = module_dependency_graph(module)
+    module_dag = _retain_edges(_module_dependency_graph(module), "dependency")
     cycles = list(nx.simple_cycles(module_dag))
     if cycles:
         messages.append(f"Cycles: {list(cycles)}")
@@ -280,6 +265,7 @@ cli.add_command(validate_module)
 @click.argument("cluster_path", type=click.Path(exists=True))
 def show_hard_dependencies_in_cluster(node_id, cluster_path: Path) -> None:
     """Show all knowledge dependencies of a given node within a particular cluster."""
+    # not worth rewriting with networkx ATM
     cluster: Cluster = Cluster.parse_file(cluster_path)
     node_in_cluster = False
     for present_node in cluster.nodes:
@@ -312,42 +298,13 @@ def get_hard_dependencies_in_module(
     module = Module(clusters=[])
     for path in cluster_paths:
         module.clusters.append(Cluster.parse_file(path))
-    node_in_module = False
-    for cluster in module.clusters:
-        for present_node in cluster.nodes:
-            present_node_id = present_node.id
-            # always explicitly namespace in module context
-            if "__" not in present_node_id:
-                present_node_id = f"{cluster.namespace_prefix}__{present_node_id}"
-            if present_node_id == node_id:
-                # could initialize several times but is harmless
-                dependencies: set[str] = set()
-                added_nodes = set([node_id])
-                node_in_module = True
-    if not node_in_module:
+    nxgraph = _retain_edges(_module_dependency_graph(module), "dependency")
+    nodes = [node for node in nxgraph.nodes if node.id == node_id]
+    if len(nodes) == 0:
         raise ValueError(
             f"Desired node {node_id} is not in the module. Did you check the namespace?"
         )
-    while len(added_nodes) > 0:
-        moved_nodes: set[str] = set(added_nodes)
-        dependencies = dependencies.union(moved_nodes)
-        added_nodes = set()
-        for moved_node in moved_nodes:
-            for cluster in module.clusters:
-                namespacify = (
-                    lambda x: x if "__" in x else f"{cluster.namespace_prefix}__{x}"
-                )
-                added_nodes = added_nodes.union(
-                    set(
-                        [
-                            namespacify(edge.start_id)
-                            for edge in cluster.edges
-                            if namespacify(edge.end_id) == moved_node
-                        ]
-                    )
-                )
-    dependencies.remove(node_id)
-    return dependencies
+    return set([node.id for node in nx.ancestors(nxgraph, nodes[0])])
 
 
 @click.command()
@@ -374,14 +331,25 @@ def check_learning_path(cluster_paths: list[Path]) -> None:
     comment_pattern = re.compile(r" *#.*")
     try:
         while True:
-            line = re.sub(comment_pattern,'',input().strip())
+            line = re.sub(comment_pattern, "", input().strip())
             if line != "":
                 learning_path.append(line)
     except EOFError:
         pass
     if len(learning_path) > 0:
-        seen_namespaced: set[str] = set()
+        complete_graph = _module_dependency_graph(module)
+        motivations_graph = _retain_edges(complete_graph, "motivation")
+        dependency_graph = _retain_edges(complete_graph, "dependency")
+        seen_nodes: set[Node] = set()
         for (index, namespaced_node_id) in enumerate(learning_path):
+            matching_nodes = [
+                node for node in complete_graph.nodes if node.id == namespaced_node_id
+            ]
+            if not matching_nodes:
+                raise ValueError(
+                    f"Node does not occur in the graph: {namespaced_node_id}"
+                )
+            node = matching_nodes[0]
             namespaced_hard_dependencies = get_hard_dependencies_in_module(
                 namespaced_node_id, cluster_paths
             )
@@ -391,65 +359,32 @@ def check_learning_path(cluster_paths: list[Path]) -> None:
                     print(f"Starting node has hard dependencies:")
                     for dependency in namespaced_hard_dependencies:
                         print(dependency)
-                motivates_something = False
-                for cluster in module.clusters:
-                    if cluster.motivations:
-                        for motivation in cluster.motivations:
-                            namespaced_start = (
-                                motivation.start_id
-                                if "__" in motivation.start_id
-                                else f"{cluster.namespace_prefix}__{motivation.start_id}"
-                            )
-                            if namespaced_start == namespaced_node_id:
-                                motivates_something = True
-                                break
-                if not motivates_something:
+                if not len(nx.descendants(motivations_graph, node)):
                     print(f"Starting node does not motivate anything.")
-            # other nodes need to be reachable given what has been visited before
-            # there also needs to be a motivation to visit them
-            # this may be because they are a necessary intermediate step
             else:
                 unmet_dependencies = namespaced_hard_dependencies.difference(
-                    seen_namespaced
+                    set((node.id for node in seen_nodes))
                 )
                 for dependency in unmet_dependencies:
                     print(
                         f"Node {index} ({namespaced_node_id}) has unmet dependency {dependency}."
                     )
-                is_motivated = False
-                # don't include motivations in graph
-                # this means the current node is a *non-optional* step towards something
-                # (everything is a non-optional step towards itself, so direct motivation is possible)
-                module_dag = module_dependency_graph(module, include_motivations=False)
-                current_and_dependents_namespaced = nx.descendants(
-                    module_dag, namespaced_node_id
+                self_and_dependents = nx.descendants(dependency_graph, node).union(
+                    set([node])
                 )
-                current_and_dependents_namespaced.add(namespaced_node_id)
-                for cluster in module.clusters:
-                    if cluster.motivations:
-                        for motivation in cluster.motivations:
-                            namespaced_start = (
-                                motivation.start_id
-                                if "__" in motivation.start_id
-                                else f"{cluster.namespace_prefix}__{motivation.start_id}"
-                            )
-                            namespaced_end = (
-                                motivation.end_id
-                                if "__" in motivation.end_id
-                                else f"{cluster.namespace_prefix}__{motivation.end_id}"
-                            )
-                            if (
-                                namespaced_end in current_and_dependents_namespaced
-                                and namespaced_start in seen_namespaced
-                            ):
-                                is_motivated = True
-                                break
+                is_motivated = False
+                for seen_node in seen_nodes:
+                    motivated_by_seen_node = nx.descendants(
+                        motivations_graph, seen_node
+                    )
+                    if len(motivated_by_seen_node.intersection(self_and_dependents)):
+                        is_motivated = True
+                        break
                 if not is_motivated:
                     print(
-                        f"Node {index} ({namespaced_node_id}) is not a motivated part of the learning path."
+                        f"Node {index} ({node.id}) is not a motivated part of the learning path."
                     )
-
-            seen_namespaced.add(namespaced_node_id)
+            seen_nodes.add(node)
     else:
         print("Empty learning path is technically valid, but pointless.")
     print("Done checking path.")
@@ -458,40 +393,118 @@ def check_learning_path(cluster_paths: list[Path]) -> None:
 cli.add_command(check_learning_path)
 
 
-def module_dependency_graph(module: Module, include_motivations=True):
+def _module_dependency_graph(module: Module):
     """Assumes the module has already been validated."""
     dag = nx.DiGraph()
     for cluster in module.clusters:
         for node in cluster.nodes:
-            dag.add_node(f"{cluster.namespace_prefix}__{node.id}")
+            inserted_node = node.copy(
+                update={"id": f"{cluster.namespace_prefix}__{node.id}"}
+            )
+            dag.add_node(inserted_node)
     for cluster in module.clusters:
         for edge in cluster.edges:
-            u_of_edge = (
+            u = (
                 edge.start_id
                 if "__" in edge.start_id
                 else f"{cluster.namespace_prefix}__{edge.start_id}"
             )
-            v_of_edge = (
+            v = (
                 edge.end_id
                 if "__" in edge.end_id
                 else f"{cluster.namespace_prefix}__{edge.end_id}"
             )
-            dag.add_edge(u_of_edge, v_of_edge)
-        if include_motivations and cluster.motivations:
+            u_node = [node for node in dag.nodes if node.id == u][0]
+            v_node = [node for node in dag.nodes if node.id == v][0]
+            dag.add_edge(u_node, v_node, kind="dependency")
+        if cluster.motivations:
             for motivation in cluster.motivations:
-                u_of_edge = (
-                    edge.start_id
-                    if "__" in edge.start_id
-                    else f"{cluster.namespace_prefix}__{edge.start_id}"
+                u = (
+                    motivation.start_id
+                    if "__" in motivation.start_id
+                    else f"{cluster.namespace_prefix}__{motivation.start_id}"
                 )
-                v_of_edge = (
-                    edge.end_id
-                    if "__" in edge.end_id
-                    else f"{cluster.namespace_prefix}__{edge.end_id}"
+                v = (
+                    motivation.end_id
+                    if "__" in motivation.end_id
+                    else f"{cluster.namespace_prefix}__{motivation.end_id}"
                 )
-                dag.add_edge(u_of_edge, v_of_edge)
+                u_node = [node for node in dag.nodes if node.id == u][0]
+                v_node = [node for node in dag.nodes if node.id == v][0]
+                dag.add_edge(u_node, v_node, kind="motivation")
     return dag
 
 
+def _retain_edges(graph, kind: Literal["dependency"] | Literal["motivation"]):
+    """Produce a new graph which only has the specified kind of edges"""
+    copy = graph.copy()
+    removed = [
+        (u, v) for (u, v) in copy.edges if copy.get_edge_data(u, v)["kind"] != kind
+    ]
+    for (u, v) in removed:
+        copy.remove_edge(u, v)
+    return copy
+
+
+@click.command()
+@click.argument("cluster_paths", type=click.Path(exists=True), nargs=-1)
+def generate_moodle_structure(cluster_paths: list[Path]) -> None:
+    """Generates JSON data for the Moodle plugin."""
+    module = Module(clusters=[])
+    for path in cluster_paths:
+        module.clusters.append(Cluster.parse_file(path))
+    dag = _module_dependency_graph(module)  # includes both kinds
+    dependency_graph = _retain_edges(dag, "dependency")
+    motivation_graph = _retain_edges(dag, "motivation")
+    nodes_in_structure: Dict = dict()
+    for node in dag:
+        nodes_in_structure[node.id] = dict()
+        if node.title:
+            nodes_in_structure[node.id]["title"] = node.title
+        if node.assignments:
+            nodes_in_structure[node.id]["assignments"] = []
+            for assignment in node.assignments:
+                list_element = {
+                    "description": assignment.description,
+                    "path": str(assignment.path),
+                }
+                if assignment.attachments:
+                    list_element["attachments"] = list((str(path) for path in assignment.attachments))
+                nodes_in_structure[node.id]["assignments"].append(list_element)
+        if node.urls:
+            nodes_in_structure[node.id]["URLs"] = []
+            for URL in node.urls:
+                nodes_in_structure[node.id]["URLs"].append(
+                    {"url": URL.url, "description": URL.description}
+                )
+    enablers: Dict = dict()
+    for node in dag.nodes():
+        self_and_dependents: Set[Node] = nx.descendants(dependency_graph, node).union(
+            set([node])
+        )
+        current_node_enablers = []
+        non_dependents = set(dag.nodes())
+        for non_dependent in non_dependents:
+            for self_or_dependent in self_and_dependents:
+                if motivation_graph.has_edge(non_dependent, self_or_dependent):
+                    current_node_enablers.append(non_dependent.id)
+        enablers[node.id] = current_node_enablers
+    structure: Dict = {
+        "nodes": nodes_in_structure,
+        # deliberately only adding *immediate* predecessors to avoid redundancy
+        "requires_all_of": {
+            node.id: [
+                predecessor.id for predecessor in dependency_graph.predecessors(node)
+            ]
+            for node in dag.nodes()
+        },
+        "motivated_by_any_of": enablers,
+    }
+    print(json.dumps(structure, sort_keys=True, indent=4))
+
+
+cli.add_command(generate_moodle_structure)
+
+
 if __name__ == "__main__":
     cli()
